/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: Locator.java,v 8142af77491e 2010/01/15 16:52:31 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.util;

import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import javax.annotation.concurrent.ThreadSafe;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.awt.EventQueue;
import org.openide.util.Lookup;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: Locator.java,v 8142af77491e 2010/01/15 16:52:31 fabrizio $
 *
 **********************************************************************************************************************/
@ThreadSafe
public final class Locator<T>
  {
    @Retention(RetentionPolicy.RUNTIME)
    public @interface ImplementedBy
      {
        Class<?> value();
      }

    public static class NotFoundException extends RuntimeException
      {
        public NotFoundException (final @Nonnull Class<?> serviceClass)
          {
            super("Not found: " + serviceClass);
          }

        public NotFoundException (final @Nonnull String name)
          {
            super("Not found: " + name);
          }
      }

    // Replacement for TopComponent.class, since we don't want a static dependency on org.openide.windows.
    @CheckForNull
    private final static Class<?> TOP_COMPONENT_CLASS;

    static
      {
        Class<?> clazz = null;

        try
          {
            final ClassLoader classLoader = Lookup.getDefault().lookup(ClassLoader.class);
            clazz = classLoader.loadClass("org.openide.windows.TopComponent");
          }
        catch (ClassNotFoundException ex)
          {
            clazz = null;
          }

        TOP_COMPONENT_CLASS = clazz;
      }

    private Locator()
      {
      }
    
    @Nonnull
    public static final <T> T find (final @Nonnull Class<T> serviceClass)
      {
        Class<? extends T> serviceClass2 = serviceClass;
        final ImplementedBy implementedBy = serviceClass.getAnnotation(ImplementedBy.class);

        if (implementedBy != null)
          {
            serviceClass2 = (Class<? extends T>)implementedBy.value();
          }

        if ((TOP_COMPONENT_CLASS != null) && TOP_COMPONENT_CLASS.isAssignableFrom(serviceClass2))
          {
            return topComponentFind(serviceClass2);
          }
        else
          {
            return simpleFind(serviceClass2);
          }
      }

    @Nonnull
    private static final <T> T simpleFind (final @Nonnull Class<T> serviceClass)
      {
        final T service = Lookup.getDefault().lookup(serviceClass);

        if (service == null)
          {
            throw new NotFoundException(serviceClass);
          }

        return service;
      }

    @Nonnull
    private static synchronized final <T> T topComponentFind (final @Nonnull Class<T> topComponentClass)
      {
        final Object[] topComponent = new Object[1];

        final Runnable runnable = new Runnable()
          {
            @Override
            public void run()
              {
                try
                  {
                    final Method method = topComponentClass.getMethod("findInstance");
                    topComponent[0] = method.invoke(null);
                  }
                catch (RuntimeException e)
                  {
                    throw new RuntimeException("Cannot lookup " + topComponentClass, e);
                  }
                catch (Exception e)
                  {
                    throw new RuntimeException("Cannot lookup " + topComponentClass, e);
                  }
              }
          };

        if (EventQueue.isDispatchThread())
          {
            runnable.run();
          }
        else
          {
            try
              {
                EventQueue.invokeAndWait(runnable);
              }
            catch (InterruptedException e)
              {
                throw new RuntimeException(e);
              }
            catch (InvocationTargetException e)
              {
                throw new RuntimeException(e);
              }
          }

        if (topComponent[0] == null)
          {
            throw new NotFoundException(topComponentClass.getSimpleName());
          }

        return (T)topComponent[0];
      }
  }
